# mvcprovider

Create a ***simple MVC*** architecture based on the ***Provider*** package.

## Getting Started

Everybody need to create a **counter module**,
and of course you need it too.

## STEP 1

First, define your very complex **model** :

```
    class MyCounterModel extends Model {

      int _count = 0;
      set count(int value) {
         _count = value;
        refresh();
       }
      int get count => _count;

    }
```

## STEP 2

Then define a **controller** that will increment the *count* property
with a lot of mathematics operations :

```
    class MyCounterCtrl extends Controller<MyCounterModel> {
    
      void increment() {
        model.count++;
      }
    
      /*
      // Uncomment if needed

      // You can access to injected providers everywhere in your controller and your view after the first view render
      // Just by using : get[MyProviderClass]

      // Init is called one time before view's first render
      // But you don't have access yet to the dependency getter: get[MyProviderClass]
      @override
      Controller init(BuildContext context, MyCounterModel model, Map<Type, dynamic> dependencies) {
        super.init(context, model, dependencies);

        // Do some cool things here that you need to do before
        // using this controller or rendering the view

        return this;
      }

      // OnReady is called one time after view's first render
      // Now you have access to the dependency getter: get[MyProviderClass] :D
      @override
      void onReady() {
        Log.print('onReady ${get[RouteService].params[RouteParams.ItemId]}', title: 'PixabayDetailsCtrl');
        model.image = get[CacheService].images[get[RouteService].params[RouteParams.ItemId]];
        refresh();
      }

      // OnRefresh is called each time the view is rendered
      @override
      Controller onRefresh(BuildContext) {
        super.onRefresh(context);

        // Do some cool things each time you refresh the view

        return this;
      }
      */
    
    }
```

For now, you have an amazing method *increment* that will increment (no ?! really ??)
you're model's property *count* which will call *refresh* (a method alias for the method *notifyListeners* inherited from her mother class **ChangeNotifier**)
in order to refresh the **view**.

## STEP 3

But to refresh the **view**. You need a **view**. Here it is :

```
    class MyCounterView extends View<MyCounterModel, MyCounterCtrl> {
    
      @override
      Widget build(BuildContext context, [Widget child]) {
        super.build(context);
    
        return Column(
          mainAxisAlignment: MainAxisAlignment.center,
          children: <Widget>[
            Text("Counter : ${model.count}"),
            RaisedButton(
                child: Text("Increment"),
                onPressed: ctrl.increment,
            )
          ]
        );
      }
    }
```

It's a bit complicated **view**. I'll explain : it's a *text* on top of a *button*.

Like you can see the *text* display our **model**'s property *count* and the
*button* execute our **controller**'s method *increment* when pressed.

## STEP 4

In order to make the magic happen you'll need to link these **three entities**
together, I decided that this fusion will be called a **Module** but you can
call this **Roberto** or anything else if you prefer.

Here is our **module** :

```
    class MyCounterModule extends Module<MyCounterModel, MyCounterView, MyCounterCtrl> {
    
      final MyCounterModel model = MyCounterModel();
      final MyCounterView view = MyCounterView();
      final MyCounterCtrl ctrl = MyCounterCtrl();
    
      /*
      // Uncomment if needed

      // You can access to injected providers everywhere in your controller and your view
      // Just by using : get[MyProviderClass]

      // You can declare & inject here all brand new providers that you need
      final List<SingleChildWidget> providers = [
        ChangeNotifierProvider(create: (context) => MyProviderClass1()),
        ChangeNotifierProvider(create: (context) => MyProviderClass2()),
      ];

      // You can inject here all already declared providers that you need
      final Map<Type, dynamic> dependencies = {
        MyProviderClass2: (context) => Provider.of<MyProviderClass1>(context)
        MyProviderClass3: (context) => Provider.of<MyProviderClass2>(context)
        MyProviderClass4: (context) => Provider.of<MyProviderClass3>(context)
      };
      */
    
    }
```

## STEP 5

At this point if you're a *good* developer you should have a *module* folder named ***mycounter***
with 4 files inside :

>The module ***my.counter.dart***
>The model ***my.counter.model.dart***
>The view ***my.counter.view.dart***
>The controller ***my.counter.ctrl.dart***

You can also have only one file with all your code inside but it's mean you're a *bad* developer ;)

To test you newly counter feature in your app just call your widget ***MyCounterModule()***
Example in your main.dart :

```
    void main() => runApp(AmazingSample());
    
    class AmazingSample extends StatelessWidget {
      // This widget is the root of this amazing package sample.
      @override
      Widget build(BuildContext context) {
        return MaterialApp(
          title: 'MVC Provider counter demo',
          home: MyCounterModule(),
        );
      }
    }
```
