library mvcprovider;

import 'dart:async';
import 'package:flutter/material.dart';
import 'package:provider/provider.dart';
import 'package:provider/single_child_widget.dart';

///
/// The module class which is the link between MVC entities
///
/// Sample :
///
/// class MyCounterModule extends Module<MyCounterModel, MyCounterView, MyCounterCtrl> {
///
///   final MyCounterModel model = MyCounterModel();
///   final MyCounterView view = MyCounterView();
///   final MyCounterCtrl ctrl = MyCounterCtrl();
/// }
class Module<M extends Model, V extends Widget, C extends Controller>
    extends StatelessWidget {
  final M model;
  final V view;
  final C ctrl;
  final List<SingleChildWidget> providers;
  final Map<Type, dynamic> dependencies;

  Module({this.model, this.view, this.ctrl, this.providers, this.dependencies});

  bool get isProviders => providers != null && providers.length > 0;

  @override
  Widget build(BuildContext context) {
    return MultiProvider(providers: [
      ...(isProviders ? providers : []),
      ChangeNotifierProvider.value(value: model),
      ChangeNotifierProvider.value(
          value: ctrl.init(context, model, dependencies))
    ], child: view);
  }
}

///
/// The Model class which will contain data and optionally
/// some methods to format this data
///
/// Sample :
///
/// class MyCounterModel extends Model {
///
///   int _count = 0;
///   set count(int value) {
///     _count = value;
///     refresh();
///   }
///   int get count => _count;
/// }
class Model extends ChangeNotifier {
  Function get refresh => notifyListeners;
  Map<Type, dynamic> get;
}

///
/// The Controller class which will handle user's actions
///
/// Sample :
///
/// class MyCounterCtrl extends Controller<MyCounterModel> {
///
///   void increment() {
///     model.count++;
///   }
/// }
class Controller<M extends Model> extends ChangeNotifier {
  BuildContext context;
  M model;
  Function get refresh => notifyListeners;
  Map<Type, dynamic> _dependencies;
  Map<Type, dynamic> get;

  bool _ready = false;
  StreamController<bool> _isReady;
  bool _listenerClosed = false;
  StreamSubscription<bool> _listener;

  // Called every rendering
  Controller onRefresh(BuildContext context) {
    this.context = context;
    this.get =
        (_dependencies ?? {}).map((type, fct) => MapEntry(type, fct(context)));
    if (!_ready) {_isReady.add(true);}
    else if (!_listenerClosed) {
      _listenerClosed = true;
      deactivate();
    }
    this.model.get = this.get;
    return this;
  }

  // Called before first rendering
  Controller init(
      BuildContext context, M model, Map<Type, dynamic> dependencies) {
    this.context = context;
    this.model = model;
    _dependencies = dependencies;
    _ready = false;
    _isReady = StreamController<bool>();
    _listener = _isReady.stream.listen((_) {
      _ready = true;
      onReady();
    });
    return this;
  }

  // Called after first rendering
  void onReady() {}

  void deactivate() {
    _listener.cancel();
    _isReady.close();
  }

  @override
  void dispose() {
    deactivate();
    super.dispose();
  }
}

///
/// A private class which consist in a type definition
/// in order to keep view's properties immutable
///
class _MC<M, C> {
  M model;
  C ctrl;
}

///
/// The View mixin which will display model data and link
/// controller's methods to user's actions
///
/// Sample :
///
/// class MyCounterView extends View<MyCounterModel, MyCounterCtrl> {
///
///   @override
///   Widget build(BuildContext context, [Widget child]) {
///     super.build(context);
///
///     return Column(
///         mainAxisAlignment: MainAxisAlignment.center,
///         children: <Widget>[
///           Text("Counter : ${model.count}"),
///           RaisedButton(
///             child: Text("Increment"),
///             onPressed: ctrl.increment,
///           )
///         ]
///     );
///   }
/// }
mixin View<M extends Model, C extends Controller> {
  final _MC<Model, C> _ = _MC<Model, C>();
  M get model => _.model;
  C get ctrl => _.ctrl;
  Map<Type, dynamic> get get => _.ctrl.get;

  void listenForChanges(BuildContext context) {
    listenForModelChanges(context);
    listenForControllerChanges(context);
  }

  void listenForModelChanges(BuildContext context) {
    _.model = Provider.of<M>(context);
  }

  void listenForControllerChanges(BuildContext context) {
    _.ctrl = Provider.of<Controller>(context).onRefresh(context);
  }
}
